/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMshCrack.h"
#include "wMshData.h"
#include "wGroup.h"
#include "wGroups.h"
#include "wTag.h"
#include "wNode.h"
#include "wElement.h"
#include "wLine2.h"
#include "wTri3.h"

using namespace tbox;

MshCrack::MshCrack(std::shared_ptr<MshData> _msh, int _nDim) : wSharedObject(), msh(_msh), nDim(_nDim)
{
    hasCrack = true; // so that we prevent automatic data structure creation if setCrack() is not called
    crckGrp = nullptr;
    exclGrp = nullptr;
}
MshCrack::~MshCrack()
{
    // Delete temporary groups
    delete crckGrp;
    delete exclGrp;
    for (size_t i = 0; i < sBndGrps.size(); ++i)
        delete sBndGrps[i];
    for (size_t i = 0; i < sBndGrp.size(); ++i)
        delete sBndGrp[i];
    for (size_t i = 0; i < vBndGrps.size(); ++i)
        delete vBndGrps[i];
    for (size_t i = 0; i < vBndGrp.size(); ++i)
        delete vBndGrp[i];
    std::cout << "~MshCrack()\n";
}

/**
 * @brief Create temporary group for crack
 */
void MshCrack::setCrack(std::string const &name)
{
    // Check if a crack needs to be made
    auto it = msh->ntags.find(name + '_');
    if (it != msh->ntags.end())
        std::cout << "Groups " << name << " and " << name << "_ are already existing. MshCrack will not be run\n";
    else
    {
        hasCrack = false;
        crckGrp = new Group(msh, name);
        if (crckGrp->tag->dim != nDim - 1)
        {
            std::stringstream err;
            err << "MshCrack::setCrack: cannot create a crack of dimension " << crckGrp->tag->dim << ", dimension must be " << nDim - 1 << "!\n";
            throw std::runtime_error(err.str());
        }
    }
}

/**
 * @brief Create temporary group for excluded nodes on crack
 */
void MshCrack::setExcluded(std::string const &name)
{
    if (!hasCrack)
        exclGrp = new Group(msh, name);
}

/**
 * @brief Create temporary groups for crack boundaries
 */
void MshCrack::addBoundary(std::string const &name)
{
    if (!hasCrack)
    {
        // Check if the boundaries on both sides of the crack are explicitely provided in two different tags
        // if so, store them
        try
        {
            Groups *gs = new Groups(msh, std::vector<std::string>{name, name + '_'});
            if (gs->groups[0]->tag->dim == nDim)
                vBndGrps.push_back(gs);
            else if (gs->groups[0]->tag->dim == nDim - 1)
                sBndGrps.push_back(gs);
            else
            {
                std::stringstream err;
                err << "MshCrack::addBoundary: cannot add a boundary of dimension " << gs->groups[0]->tag->dim << ", dimension must be either " << nDim << " or " << nDim - 1 << "!\n";
                throw std::runtime_error(err.str());
            }
        }
        // else, the boundaries are contained in a single tag, an algorithm will later try to separate the elements according to the side of the crack they are on
        catch (const std::out_of_range &)
        {
            Group *g = new Group(msh, name);
            if (g->tag->dim == nDim)
                vBndGrp.push_back(g);
            else if (g->tag->dim == nDim - 1)
                sBndGrp.push_back(g);
            else
            {
                std::stringstream err;
                err << "MshCrack::addBoundary: cannot add a boundary of dimension " << g->tag->dim << ", dimension must be either " << nDim << " or " << nDim - 1 << "!\n";
                throw std::runtime_error(err.str());
            }
        }
    }
}

/**
 * @brief Create the crack and save the mesh
 */
void MshCrack::run()
{
    if (hasCrack)
        std::cout << "MshCrack not run\n";
    else
    {
        // Check that crack is bounded
        if ((sBndGrps.empty() && sBndGrp.empty()) || (vBndGrps.empty() && vBndGrp.empty()))
            throw std::runtime_error("MshCrack::run, crack has no (surface and/or volume) boundary, use MshCrack::addBoundary()!\n");
        // Open the crack and modify data structure
        openCrack();
        modifyBoundaries();
        std::cout << "Crack added on " << *(crckGrp->tag) << "(" << crckGrp->tag->name << "_ created)" << std::endl;
    }
}

/**
 * @brief Duplicate the nodes and elements on the crack and modify the mesh accordingly
 */
void MshCrack::openCrack()
{
    // Collect the nodes to be duplicated
    std::vector<Node *> nodes;
    for (auto e : crckGrp->tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);
    std::sort(nodes.begin(), nodes.end());
    auto it = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), it));
    // Remove the nodes to be excluded
    if (exclGrp != nullptr)
    {
        // Get junction nodes
        std::vector<Node *> excl;
        for (auto e : exclGrp->tag->elems)
            for (auto n : e->nodes)
                excl.push_back(n);
        std::sort(excl.begin(), excl.end());
        it = std::unique(excl.begin(), excl.end());
        excl.resize(std::distance(excl.begin(), it));
        // Find the intersection and remove
        std::vector<Node *>::iterator endRange;
        endRange = std::set_difference(nodes.begin(), nodes.end(),
                                       excl.begin(), excl.end(),
                                       nodes.begin());
        nodes.erase(endRange, nodes.end());
    }
    // Duplicate the nodes, modify the mesh and make a map
    size_t i = msh->nodes.size();
    for (auto n : nodes)
    {
        Node *newNodep = new Node(static_cast<int>(i + 1), n->pos, static_cast<int>(i));
        msh->nodes.push_back(newNodep);
        nodMap[n] = newNodep;
        i++;
    }

    // Create new tag and modify the mesh
    Tag *tagp = new Tag(static_cast<int>(msh->ptags.rbegin()->first + 1), crckGrp->tag->name + "_", nDim - 1);
    Tag *tage = new Tag(static_cast<int>(msh->etags.rbegin()->first + 1), crckGrp->tag->name + "_", nDim - 1);
    msh->ntags[crckGrp->tag->name + "_"] = tagp;
    msh->ptags[static_cast<int>(msh->ptags.rbegin()->first + 1)] = tagp;
    msh->etags[static_cast<int>(msh->etags.rbegin()->first + 1)] = tage;
    // Duplicate elements and modify the mesh/tag
    i = msh->elems.size() + 1;
    for (auto e : crckGrp->tag->elems)
    {
        std::vector<Node *> nods(e->nodes.size());
        for (size_t j = 0; j < nods.size(); ++j)
        {
            try
            {
                nods[j] = nodMap.at(e->nodes[e->nodes.size() - j - 1]);
            }
            catch (const std::out_of_range &)
            {
                nods[j] = e->nodes[e->nodes.size() - j - 1];
            }
        }
        if (e->type() == ElType::LINE2 && nDim == 2)
            msh->elems.push_back(new Line2(static_cast<int>(i), tagp, tage, e->parts, nods));
        else if (e->type() == ElType::TRI3 && nDim == 3)
            msh->elems.push_back(new Tri3(static_cast<int>(i), tagp, tage, e->parts, nods));
        else
        {
            std::stringstream err;
            err << "MshCrack::run: could not create element of type " << e->type() << " for dimension " << nDim << "!\n";
            throw std::runtime_error(err.str());
        }
        i++;
    }
}

/**
 * @brief Modify the elements on the negative side of the crack
 *
 * 1) Identify boundary surface elements on negative side of crack, and store their nodes
 *   - easy case: elements are contained in different tags
 *   - hard case: elements are contained in a single tag, they are identified by comparing the vector joining them to the crack to the normal vector of the crack (NOT ROBUST)
 * 2) Identify boundary volume elements on negative side of crack
 *   - easy case: elements are contained in different tags
 *   - hard case: elements are contained in a single tag, they are identified by
 *     - either checking if one of their nodes belong to the surface
 *     - or, comparing the vector joining them to the crack to the normal vector of the crack
 * 3) Swap the nodes of (surface and volume) elements located on the negative side of the crack
 * 4) Merge the tags identifying the (surface and volume) elements on the negative side into the tags identifying the positive side
 */
void MshCrack::modifyBoundaries()
{
    // Store crack nodes, and compute CGs and normals if sides are contained in a single tag
    std::unordered_set<Node *> crkNods, crkNodsAll;
    std::vector<Eigen::Vector3d> crkCG(crckGrp->tag->elems.size()), crkN(crckGrp->tag->elems.size());
    if (!sBndGrp.empty() || !vBndGrp.empty())
    {
        // nodes of crack except excluded
        for (auto p : nodMap)
            crkNods.insert(p.first);
        // all nodes of crack
        for (auto e : crckGrp->tag->elems)
            crkNodsAll.insert(e->nodes.begin(), e->nodes.end());
        // cg and unit normal vector
        for (size_t i = 0; i < crckGrp->tag->elems.size(); ++i)
        {
            crckGrp->tag->elems[i]->computeCg();
            crkCG[i] = crckGrp->tag->elems[i]->cg;
            crckGrp->tag->elems[i]->computeNormal();
            crkN[i] = crckGrp->tag->elems[i]->normal;
        }
    }

    // Identify surface elements on negative side, and store their nodes (to find volume elements if they are contained in a single tag)
    std::vector<Element *> _surEl;
    std::unordered_set<Node *> _surNd;
    // Each side is contained in a different tag
    for (auto gs : sBndGrps)
    {
        for (auto e : gs->groups[1]->tag->elems)
        {
            _surEl.push_back(e);
            // only keep nodes that are not on crack (since they are on positive side)
            for (auto n : e->nodes)
                if (!crkNodsAll.count(n))
                    _surNd.insert(n);
        }
    }
    // Both sides are contained in a single tag
    for (auto g : sBndGrp)
    {
        // find elements having at least one node on the crack, and store if on negative side
        std::unordered_set<Element *> onCrk = findTouching(g->tag->elems, crkNods);
        for (auto e : onCrk)
        {
            // find mininal distance bewteen CGs
            e->computeCg();
            size_t idx = findClosest(crkCG, e->cg);
            // if dot product between crack-to-volume vector and crack normal is negative, element is on negative side
            if ((e->cg - crkCG[idx]).dot(crkN[idx]) < 0)
            {
                _surEl.push_back(e);
                // only keep nodes that are not on crack (since they are on positive side)
                for (auto n : e->nodes)
                    if (!crkNodsAll.count(n))
                        _surNd.insert(n);
            }
        }
    }

    // Identify volume elements touching surface elements on negative side
    std::vector<Element *> _volEl;
    // Each side is contained in a different tag
    for (auto gs : vBndGrps)
        for (auto e : gs->groups[1]->tag->elems)
            _volEl.push_back(e);
    // Both sides are contained in a single tag
    for (auto g : vBndGrp)
    {
        // find elements having at least one node on the crack
        std::unordered_set<Element *> onCrkOnly = findTouching(g->tag->elems, crkNods);
        // find elements having at least one node on the negative side of the surface
        std::unordered_set<Element *> onSurf = findTouching(std::vector<Element *>(onCrkOnly.begin(), onCrkOnly.end()), _surNd);
        // remove these elements from the first set, as they are necessarily on the negative side
        _volEl.insert(_volEl.end(), onSurf.begin(), onSurf.end());
        for (auto e : onSurf)
            onCrkOnly.erase(e);
        // check if remaining elements are on the negative side
        for (auto e : onCrkOnly)
        {
            // find mininal distance bewteen CGs
            e->computeCg();
            size_t idx = findClosest(crkCG, e->cg);
            // if dot product between crack-to-volume vector and crack normal is negative, element is on negative side
            if ((e->cg - crkCG[idx]).dot(crkN[idx]) < 0)
                _volEl.push_back(e);
        }
    }

    // Swap nodes of surface and volume elements on negative side
    swapNodes(_surEl);
    swapNodes(_volEl);
    // Merge tag identifying negative side into positive side tag
    mergeTags(sBndGrps);
    mergeTags(vBndGrps);
}

/**
 * @brief Merge physical tags and modify data structure accordingly
 */
void MshCrack::mergeTags(std::vector<Groups *> &grps)
{
    for (auto gs : grps)
    {
        // Merge tags
        Tag *refTag = gs->groups[0]->tag;
        Tag *oldTag = gs->groups[1]->tag;
        // Modify element tags
        for (auto e : gs->groups[1]->tag->elems)
            e->ptag = refTag;
        // Append elements to new tag
        refTag->elems.insert(refTag->elems.end(), oldTag->elems.begin(), oldTag->elems.end());
        // Delete old tag and remove from mesh maps
        msh->ptags.erase(oldTag->no);
        msh->ntags.erase(oldTag->name);
        delete oldTag;
    }
}

/**
 * @brief Find elements touching a surface (ie, that have at least a node in the provided set)
 */
std::unordered_set<Element *> MshCrack::findTouching(std::vector<Element *> const &elms, std::unordered_set<Node *> const &nods)
{
    std::unordered_set<Element *> found;
    for (auto e : elms)
    {
        for (auto n : e->nodes)
        {
            if (nods.count(n))
            {
                found.insert(e);
                break;
            }
        }
    }
    return found;
}

/**
 * @brief Find closest element and returns its index
 */
size_t MshCrack::findClosest(std::vector<Eigen::Vector3d> const &cga, Eigen::Vector3d const &cgb)
{
    size_t idx = 0;
    std::vector<Eigen::Vector3d> delta(cga.size());
    for (size_t i = 0; i < cga.size(); ++i)
    {
        delta[i] = cgb - cga[i];
        if (delta[i].norm() < delta[idx].norm())
            idx = i;
    }
    return idx;
}

/**
 * @brief Replace nodes of boundary elements on negative side of crack by negative crack nodes
 */
void MshCrack::swapNodes(std::vector<Element *> &elms)
{
    for (auto e : elms)
    {
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            try
            {
                e->nodes[i] = nodMap.at(e->nodes[i]);
            }
            catch (const std::out_of_range &)
            {
                // std::cout << e->nodes[i]->no << "not found in map!\n";
            }
        }
    }
}

void MshCrack::write(std::ostream &out) const
{
    out << "MshCrack on " << msh << std::endl;
}
