/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wUnitTest.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wGmshImport.h"
#include "wGmshExport.h"
#include "wResults.h"
#include "wMshDeform.h"
#include <Eigen/Dense>

using namespace tbox;

/**
 * @brief Write and read results to/from disk
 */
void UnitTest::gmshIO(GmshExport *writer, GmshImport *reader)
{
    // Mesh
    auto mesh = writer->msh; // should be the same than in reader->msh

    // Variables to be written then read
    std::vector<double> x(mesh->nodes.size(), 0), type(mesh->elems.size(), 0);     // x-position, element type
    std::vector<Eigen::Vector3d> xyz(mesh->nodes.size(), Eigen::Vector3d::Zero()); // x,y,z-position

    // Temporary results container
    Results res;
    res.scalars_at_nodes["x"] = &x;
    res.vectors_at_nodes["xyz"] = &xyz;
    res.scalars_at_elems["type"] = &type;

    // Fill the variables and write them to disk
    for (size_t i = 0; i < mesh->nodes.size(); ++i)
    {
        x[i] = mesh->nodes[i]->pos(0);
        xyz[i] = mesh->nodes[i]->pos;
    }
    for (size_t i = 0; i < mesh->elems.size(); ++i)
        type[i] = static_cast<double>(mesh->elems[i]->type());
    writer->binary = false; // because we cannot read files in bin format yet...
    writer->save(mesh->name, res);

    // Clear the variables and read them from disk
    std::fill(x.begin(), x.end(), 0);
    std::fill(xyz.begin(), xyz.end(), Eigen::Vector3d::Zero());
    std::fill(type.begin(), type.end(), 0);
    reader->load(mesh->name + ".pos", res, 0);

    // Test that the variables have the required value
    for (size_t i = 0; i < x.size(); ++i)
        if (x[i] != mesh->nodes[i]->pos(0))
            this->error("gmshIO", std::to_string(x[i]), std::to_string(mesh->nodes[i]->pos(0)));
    for (size_t i = 0; i < xyz.size(); ++i)
        if (xyz[i] != mesh->nodes[i]->pos)
            this->error("gmshIO",
                        "(" + std::to_string(xyz[i](0)) + "," + std::to_string(xyz[i](1)) + "," + std::to_string(xyz[i](2)) + ")",
                        "(" + std::to_string(mesh->nodes[i]->pos(0)) + "," + std::to_string(mesh->nodes[i]->pos(1)) + "," + std::to_string(mesh->nodes[i]->pos(2)) + ")");
    for (size_t i = 0; i < type.size(); ++i)
        if (type[i] != static_cast<double>(mesh->elems[i]->type()))
            this->error("gmshIO", std::to_string(type[i]), std::to_string(static_cast<double>(mesh->elems[i]->type())));
}

/**
 * @brief Deform a mesh of dimension d
 */
void UnitTest::mshDeform(MshDeform *mshDef, int d)
{
    // Initialize memory (volume element only)
    for (auto e : mshDef->msh->elems)
        if ((d == 2 && e->type() == ElType::TRI3) || (d == 3 && e->type() == ElType::TETRA4))
            e->initValues(true);
    // Call deform
    mshDef->deform();
}

/**
 * @brief Print error on failure
 */
void UnitTest::error(std::string const &name, std::string const &computed, std::string const &expected)
{
    std::string message = "tbox::UnitTest::" + name + " failed! Computed: " + computed + " but expected: " + expected + ".\n";
    throw std::runtime_error(message);
}
